

/**
 * Module dependencies.
 */

var request = require('superagent')
  , util = require('util')
  , http = require('http')
  , Request = request.Request;

/**
 * Starting port.
 */

var port = 3456;

/**
 * Expose `Test`.
 */

module.exports = Test;

/**
 * Initialize a new `Test` with the given `app`,
 * request `method` and `path`.
 *
 * @param {Server} app
 * @param {String} method
 * @param {String} path
 * @api public
 */

function Test(app, method, path) {
  Request.call(this, method, path);
  this.redirects(0);
  this.app = app;
  this._fields = {};
  var addr = app.address();
  var portno = addr ? addr.port : port++;
  if (!addr) app.listen(portno);
  this.url = 'http://0.0.0.0:' + portno + path;
}

/**
 * Inherits from `Request.prototype`.
 */

Test.prototype.__proto__ = Request.prototype;

/**
 * Expectations:
 *
 *   .expect(200)
 *   .expect(200, fn)
 *   .expect('Some body')
 *   .expect('Some body', fn)
 *   .expect('Content-Type', 'application/json')
 *   .expect('Content-Type', 'application/json', fn)
 *
 * @return {Test}
 * @api public
 */

Test.prototype.expect = function(val, fn){
  var self = this;

  // status
  if ('number' == typeof val) {
    this._status = val;
  }

  // body
  if ('string' == typeof val || val instanceof RegExp) {
    // header field
    if ('string' == typeof fn || fn instanceof RegExp) {
      this._fields[val] = fn;
      fn = arguments[2];
    } else {
      this._body = val;
    }
  }

  // callback
  if ('function' == typeof fn) this.end(fn);

  return this;
};

/**
 * Defer invoking superagent's `.end()` until
 * the server is listening.
 *
 * @param {Function} fn
 * @api public
 */

Test.prototype.end = function(fn){
  var self = this;
  var end = Request.prototype.end;
  end.call(this, function(res){
    self.assert(res, fn);
  });
};

/**
 * Perform assertions and invoke `fn(err)`.
 *
 * @param {Response} res
 * @param {Function} fn
 * @api private
 */

Test.prototype.assert = function(res, fn){
  var status = this._status
    , fields = this._fields
    , body = this._body
    , expected
    , actual
    , re;

  // status
  if (status && res.status !== status) {
    var a = http.STATUS_CODES[status];
    var b = http.STATUS_CODES[res.status];
    return fn(new Error('expected ' + status + ' "' + a + '", got ' + res.status + ' "' + b + '"'), res);
  }

  // body
  if (null != body && body !== res.text) {
    var a = util.inspect(body);
    var b = util.inspect(res.text);

    // regexp
    if (body instanceof RegExp) {
      if (!body.test(res.text)) {
        return fn(new Error('expected body ' + b + ' to match ' + body));
      }
    } else {
      return fn(new Error('expected ' + a + ' response body, got ' + b));
    }
  }

  // fields
  for (var field in fields) {
    expected = fields[field];
    actual = res.header[field.toLowerCase()];
    if (null == actual) return fn(new Error('expected "' + field + '" header field'));
    if (expected == actual) continue;
    if (expected instanceof RegExp) re = expected;
    if (re && re.test(actual)) continue;
    if (re) return fn(new Error('expected "' + field + '" matching ' + expected + ', got "' + actual + '"'));
    return fn(new Error('expected "' + field + '" of "' + expected + '", got "' + actual + '"'));
  }

  fn(null, res);
};

